ScenarioFramework = import('/lua/ScenarioFramework.lua')
SimQuery = import('/lua/SimPlayerQuery.lua')

-- Multiplayer Save global functions
msscript('code/sim/globals.lua')

-- this variable maps the old unit id to the newly created unit so orders can be restored
units = {}
-- this variable maps the old prop id to the newly created one so reclaim orders can be restored
props = {}
-- this varible maps transported units to their carrier so they can all be regrouped in one go
transports = {}

-- deletes globals to avoid conflicts and return memory
function ClearGlobals()
	units, props, transports = {}, {}, {}
end

-- Retreives newly created units using the id they had when they were saved
function GetUnitFromOldId( old_id )
	return unpack(units[old_id])
end

-- Retrieves the x,y,z coords of a unit using its old id
function GetPosFromOldId( old_id )
	local data = units[old_id][2]
	return Vector(data[2], data[3], data[4])
end

-- Save data for map/scenario
function SaveScenario()
	return {}
end

--[[
This function returns a lightweight table of the entities' data suitable for recreating its most important properties
]]
function SaveEntity(bpid, army, e)
	--bigdump(e)
	-- id
	local id = safecall('SAVE: Bad entity', e.GetEntityId, e)
	if not id then return end
	-- position
	local x, y, z = unpack( e:GetPosition() )
	-- heading (radians -> degrees)
	local heading = e:GetHeading()* 180/math.pi
	--local heading = e:GetOrientation()
	-- join data into table
	local data = {tonumber(id), x, y, z, heading}
	
	-- get unit data
	if e.IsUnitState ~= nil then
		-- this is a unit
		-- unit layer (air,land,sea,sub,etc)
		local layer = GetLayerId( e:GetCurrentLayer() )
		-- unit health
		local health = e:GetHealth()
		-- unit experience level
		local xp = e.VeteranLevel
		-- fire state (how the unit assigns attack priorities)
		local fire_state = e:GetFireState()
		-- state (what the unit is doing)
		local state = 0
		-- focus (entity we are helping, attacking or following or location we move towards)
		local focus = 0
		-- data specific to special units and circumstances
		local misc = {}
		-- check if paused
		if e:IsPaused() then
			misc.paused=1
		end
		-- get factory data
		if EntityCategoryContains(categories.FACTORY, e) then
			misc.rally = e:GetRallyPoint()
		end
		-- get stored missiles
		local nukeammo = e:GetNukeSiloAmmoCount()
		if nukeammo > 0 then
			misc.nukeammo = nukeammo
		end
		local tacammo = e:GetTacticalSiloAmmoCount()
		if tacammo > 0 then
			misc.tacammo = tacammo
		end
		-- get fuel
		if EntityCategoryContains(categories.AIR*categories.MOBILE, e) then
			misc.fuel = e:GetFuelRatio()
		end
		
		-- get information on what the unit is currently doing
		-- order is important because for simplicity we only allow one state per unit
		if e:IsDead() then
			-- Ignore deadites
		elseif e:IsUnitState('Enhancing') then
			state = GetStateId('Enhancing')
			-- CommandData is put here by custom ScriptTask.OnCreate
			misc.enhancement = e.commandData.Enhancement 
			misc.progress = e:GetWorkProgress()

		elseif e:IsUnitState('SiloBuildingAmmo') then
			state = GetStateId('SiloBuildingAmmo')
			misc.progress = e:GetWorkProgress()
			
		elseif e:IsUnitState('Upgrading') then
			state = GetStateId('Upgrading')
			focus = e.UnitBeingBuilt:GetEntityId()
			
		elseif e:IsUnitState('Guarding') then
			local guarding = e:GetGuardedUnit()
			if guarding then
				--LOG('guarding '..DumpUnit(guarding))
				state = GetStateId('Guarding')
				focus = guarding:GetEntityId()
			end
			
		elseif e:IsUnitState('Building') then
			local building = e.UnitBeingBuilt
			if building then
				--LOG('building '..DumpUnit(building))
				state = GetStateId('Building')
				focus = building:GetEntityId()
			end
			
		elseif e:IsUnitState('BeingBuilt') then
			local builder = e:GetParent()
			if builder then
				--LOG('built by '..DumpUnit(builder))
				state = GetStateId('BeingBuilt')
				focus = builder:GetEntityId()
			end
			
		elseif e:IsUnitState('Attached') then
			local parent = e:GetParent()
			if parent then
				--LOG('attached to '..DumpUnit(parent))
				state = GetStateId('Attached')
				focus = parent:GetEntityId()
			end
			
		elseif e:IsUnitState('Patrolling') then
			local nav = e:GetNavigator()
			if nav then
				state = GetStateId('Patrolling')
				focus = nav:GetGoalPos()
			end
			
		elseif e:IsUnitState('Repairing') then
			local repairee = e:GetFocusUnit()
			if repairee then
				--LOG('repairee '..DumpUnit(repairee))
				state = GetStateId('Repairing')
				focus = repairee:GetEntityId()
			end
			
		elseif e:IsUnitState('Attacking') then
			local target = e:GetTargetEntity()
			if target then
				--LOG('target '..DumpUnit(target))
				state = GetStateId('Attacking')
				focus = target:GetEntityId()
			end
			
		elseif e:IsUnitState('Moving') then
			local nav = e:GetNavigator()
			if nav then
				state = GetStateId('Moving')
				focus = nav:GetGoalPos()
			end
		end
		-- join entity and unit data
		data = table.join(data,{layer,health,xp,fire_state,state,focus})
		-- add optional data
		if not table.empty(misc) then
			table.insert(data,misc)
		end
		-- log
		if state ~= 0 then
			--LOG('SAVE: Unit '..id..' is '..GetStateName(state))
			-- get command queue
			--LOG(repr(e:PrintCommandQueue(),1,1))
		end
	end
	-- end unit data
	return data
end

function GetAllEntities()
	-- get data for all entities on the map
	local map_w, map_h = unpack(ScenarioInfo.size)
	return GetEntitiesInRect( 0, 0, map_w, map_h )
end

--[[
This is one of the main multiplayer save functions. Its job is gather as much data as possible on entities in the game so they can be recreated later. Not all data can be stored as it is either locked into the game engine or can't be recreated or isn't worth the effort.
]]
function SaveEntities() 
	-- get data for all entities on the map
	local entities = GetAllEntities()
	LOG('SAVE/LOAD: '..table.getn(entities)..' entities found.')
	local store = {}
	local count = 0
	local errors = 0
	for _,e in entities do
		-- type of entity
		local bpid = e:GetBlueprint().BlueprintId
		-- ignore landscaping
		if string.sub(bpid,1,5) ~= '/env/' then
			-- limit entities saved (for testing)
			count = count + 1
			if count > 100000 or errors > 5 then break end
			-- army index
			local army = e:GetArmy()
			-- get entity data
			local data = safecall('SAVE: Saving '..bpid, SaveEntity, bpid, army, e) 
			if data then
				-- add entity data to its place in the data tree
				-- organising the tree this way makes it smaller
				store[bpid] = store[bpid] or {}
				store[bpid][army] = store[bpid][army] or {}
				table.insert(store[bpid][army], data)
			else
				errors = errors + 1
			end
		end
	end
	return store
end

-- Removes all entities except environment from the game
function ClearAllEntities( )
	-- get data for all entities on the map
	local entities = GetAllEntities()
	for _,e in entities do
		-- type of entity
		local bpid = e:GetBlueprint().BlueprintId		
		-- ignore landscaping
		if string.sub(bpid,1,5) ~= '/env/' then
			-- trash the unit
			e:Destroy()
		end
	end
end

-- Create an entity using save data
function CreateProp( bpid, army, ent_data )
	--LOG('LOAD: Creating '..bpid..' for army '..army)
	-- get variables from table
	local id,x,y,z,heading = unpack(ent_data)
	-- Create the entity with the correct position and heading
	local e = CreatePropHPR(bpid, x, y, z, heading, 0, 0)
	-- rotate to correct heading
	--e:SetOrientation(heading, true) -- orientation, immediately
	-- map its old id to its new object for restoring orders
	props[id] = {e, ent_data}
end

-- Create a unit using save data
function CreateUnit( bpid, army, ent_data )
	--LOG('LOAD: Creating '..bpid..' for army '..army)
	-- get variables from table
	local id,x,y,z,heading,layer,health,xp,state,focus,misc = unpack(ent_data)
	-- move units built by factories out of the way so the factory doesn't sit on them
	if state == GetStateId('BeingBuilt') then x, z = 0, 0 end
	-- create the entity with the correct position and heading (in degrees around y)
	local e = CreateUnit2(bpid, army, GetLayerName(layer), x, z, heading)
	-- rotate to correct roll, pitch and yaw
	--e:SetOrientation(heading, true) -- orientation, immediately
	-- map unit to its old id for restoring orders
	units[id] = {e, ent_data}
end

-- Creates entities using save data
function CreateEntities( entity_data ) 
	-- loop over blueprints
	for bpid, army_list in entity_data do
		-- loop over armies using this bp
		for army, ent_list in army_list do
			-- loop over instances of this bp in this army 
			for _, ent_data in ent_list do
				-- convert number id back to string
				ent_data[1] = tostring(ent_data[1])
				-- create prop (trees, rocks, etc..) or unit
				if army == -1 or string.sub(bpid,1,7) == '/props/' then
					-- A prop
					safecall( 'CreateEntities: Couldn\'t Create prop '..bpid..' data: '..repr(ent_data),
						CreateProp, bpid, army, ent_data )
				else
					-- A unit
					safecall( 'CreateEntities: Couldn\'t Create unit '..bpid..' data: '..repr(ent_data),
						CreateUnit, bpid, army, ent_data  )
				end
			end
		end
	end
end

-- Restore unit properties
function RestoreUnitState( old_id, unit_and_data ) 
	
	-- details of unit
	local unit, unit_data = unpack(unit_and_data)
	local id,x,y,z,heading,layer,health,xp,fire_state,state_id,focus,misc = unpack(unit_data)
	local new_id = unit:GetEntityId()
	local unit_bpid = unit:GetBlueprint().BlueprintId
	local unit_label = unit_bpid..' '..old_id..' (now '..new_id..')'
	unit.unitLabel = unit_label
	
	-- restore experience
	unit:SetVeterancy(xp)
	-- restore its health
	unit:SetHealth( unit, math.min(health, unit:GetMaxHealth()) )
	-- restore fire state
	unit:SetFireState(fire_state)
	-- restore fuel
	if misc.fuel then
		unit:SetFuelRatio(misc.fuel)
	end
	-- restore toggles
	if misc.paused then
		unit:SetPaused(true)
	end
	-- restore rally points
	if misc.rally then
		IssueFactoryRallyPoint({unit}, misc.rally)
	end
	-- restore silo ammo
	if misc.nukeammo then
		unit:GiveNukeSiloAmmo(misc.nukeammo)
	end
	if misc.tacammo then
		unit:GiveTacticalSiloAmmo(misc.tacammo)
	end
	
	--LOG('state_id:',repr(state_id),repr(unit_data))
	if state_id ~= 0 then
		local state = GetStateName( state_id )
		--LOG( 'LOAD: Restoring state '..state..' to '..unit_label  )

		-- details of focus unit (the target,guard,builder,etc..)
		--LOG('focus: '..tostring(focus)..' ('..type(focus)..')')
		local focus_unit, focus_data, focus_new_id, focus_unit_bpid, focus_label = nil
		if type(focus) == 'table' then
			-- focus is a location {x,y,z}
			--LOG( 'LOAD: '..unit_label..' focused on '..repr(focus) )
		elseif focus ~= 0 then
			-- focus is a unit
			focus_unit, focus_data = GetUnitFromOldId( focus )
			focus_new_id = focus_unit:GetEntityId()
			focus_unit_bpid = focus_unit:GetBlueprint().BlueprintId
			focus_label = focus_unit_bpid..' '..focus..' (now '..new_id..')'
			--LOG( 'LOAD: '..unit_label..' focused on '..focus_label )
		end
		
		-- ENHANCING: Restore progress on enhancements
		if state == 'Enhancing' then
			--LOG( 'LOAD: '..unit_label..' is enhancing with '..repr(misc.enhancement) )
			-- store progress in the unit so the ScriptTask can use it
			unit.saveProgress = misc.progress
			IssueScript({unit}, {TaskName="EnhanceTask", Enhancement=misc.enhancement}) -- units, script
			--unit:SetWorkProgress(misc.progress)

		-- BUILDING MISSILES: Restore progress on nukes and tactical missiles
		elseif state == 'SiloBuildingAmmo' then
			--LOG( 'LOAD: '..unit_label..' is building missiles' )
			unit.missileProgress = misc.progress
			if EntityCategoryContains(categories.NUKE, unit) then
				IssueSiloBuildNuke({unit})
			else
				IssueSiloBuildTactical({unit})
			end
			
		-- UPGRADING: Restore upgrades
		elseif state == 'Upgrading' then
			if EntityCategoryContains(categories.FACTORY, focus_unit) then
				--LOG( 'LOAD: Factory '..unit_label..' is upgrading to '..focus_label )
				-- focusHealth will be used by Unit.OnStartBuild to restore the original progress 
				unit.focusHealth = focus_unit:GetHealth()
				unit.focusLabel = focus_label
				IssueUpgrade({unit}, focus_unit_bpid) -- units, blueprint
			else
				--LOG( 'LOAD: '..unit_label..' is upgrading to '..focus_label )
				unit.focusHealth = focus_unit:GetHealth()
				unit.focusLabel = focus_label
				IssueUpgrade({unit}, focus_unit_bpid) -- units, blueprint
			end
			
		-- BUILDING: Restart build
		elseif state == 'Building' then
			-- make sure parent is really a builder
			if EntityCategoryContains(categories.FACTORY, unit) then
				if EntityCategoryContains(categories.FACTORY, focus_unit) then
					--LOG( 'LOAD: Factory '..unit_label..' is building itself' )
				else
					--LOG( 'LOAD: Factory '..unit_label..' is building '..focus_label )
				end
				-- replace current unit with freshly built one
				unit.focusHealth = focus_unit:GetHealth()
				unit.focusLabel = focus_label
				IssueBuildFactory({unit}, focus_unit_bpid, 1) -- factories, blueprint, count
				--unit:SetWorkProgress(0.8)
				--focus_unit:Destroy()
				
			elseif EntityCategoryContains(categories.CONSTRUCTION, unit) then
				--LOG( 'LOAD: Builder '..focus_label..' is building '..unit_label )
				local build_pos = GetPosFromOldId(focus)
				local huh = {}
				--LOG('LOAD: IssueBuildMobile({unit}, '..repr(build_pos)..', '..repr(focus_unit_bpid)..', '..repr(huh)..')')
				unit.focusHealth = focus_unit:GetHealth()
				unit.focusLabel = focus_label
				IssueBuildMobile({unit}, build_pos, focus_unit_bpid, huh)
				--unit:SetWorkProgress(0.8)
			end
		
		-- ATTACHED: Reattach transported units
		elseif state == 'Attached' then
			local parent = GetUnitFromOldId( focus )
			-- check parent is definitely a transport
			if EntityCategoryContains(categories.TRANSPORTATION, focus_unit) then
				--LOG( 'LOAD: Attaching '..unit_label..' to '..focus_label ) 
				-- Add both units to global transport list
				transports[focus_new_id] = transports[focus_new_id] or {}
				table.insert(transports[focus_new_id], unit)
			end

		-- GUARDING: Continue guard
		elseif state == 'Guarding' then
			--LOG( 'LOAD: '..unit_label..' guarding '..focus_label ) 
			IssueGuard({unit}, focus_unit) -- units, guarded unit
		
		-- REPAIRING: Continue repair
		elseif state == 'Repairing' then
			--LOG( 'LOAD: '..unit_label..' repairing '..focus_label ) 
			IssueRepair({unit}, focus_unit) -- units, unit being repaired
		
		-- ATTACKING: Restart attacks
		elseif state == 'Attacking' then
			--LOG( 'LOAD: '..unit_label..' attacking '..focus_label ) 
			IssueAttack({unit}, focus_unit) -- attackers, target

		-- PATROL: Continue patrol
		-- TODO: Find out how to get all patrol points
		elseif state == 'Patrolling' then
			--LOG( 'LOAD: '..unit_label..' patrolling to '..repr(focus) ) 
			IssuePatrol({unit}, focus) -- units, location
			-- return to current position
			IssuePatrol({unit}, unit:GetPosition()) -- units, location
		
		-- MOVING: Continue move
		elseif state == 'Moving' then
			--LOG( 'LOAD: '..unit_label..' moving to '..repr(focus) ) 
			IssueMove({unit}, focus) -- units, location
		end

	end
end

-- Reattach units to a transport
function AttachTransportedUnits( transport_id, units )
	for transport_id, units in transports do
		--LOG( 'LOAD: Loading transport '..transport_id..' with '..table.getn(units)..' units' )
		ScenarioFramework.AttachUnitsToTransports(units, {GetUnitById(transport_id)})
	end
end

-- Save army properties
function SaveArmy(army_name)		
	local army_data = {}
	local army = ScenarioInfo.ArmySetup[army_name]
	local brain = GetArmyBrain(army_name)
	army_data.index = brain:GetArmyIndex()
	-- economy
	army_data.mass = brain:GetEconomyStored( 'Mass' )
	army_data.energy = brain:GetEconomyStored( 'Energy' )
	--LOG(repr(army_data))
	return army_data
end

-- Save properties for all armies
function SaveArmies()
	local armies_data = {}
	for army_name in ListArmies() do
		armies_data[army_name] = safecall( 'Saving army '..army_name, SaveArmy, army_name )
	end
	return armies_data
end

-- Restore army properties
function RestoreArmy(army_name, army_data)
	local brain = GetArmyBrain(army_name)
	SetArmyEconomy( army_name, army_data.mass, army_data.energy )
	-- Restart AI (it gets stuck on load)
	-- InitializeArmyAI(army_name)
end

-- Complete save process

-- query data is passed from the User scope
function SaveGame( query_data )
	LOG('SAVE: Sim side of save running with options: '..repr(query_data))
	local save_data = query_data.Data or {}
	-- reset globals to prevent conflicts with previous saves
	ClearGlobals()
	-- save map/scenario details
	save_data.scenario_data = SaveScenario()
	-- save entities
	save_data.entity_data = safecall( 'Saving entities '..repr(query_data), SaveEntities )
	-- save army data
	save_data.armies_data = SaveArmies()
	-- Return compressed game data to UI side
	query_data.Data = table.serialise( save_data )
	Sync.SaveGameData = query_data
end

-- Complete loading process
-- save data is passed from the User scope
function LoadGame( query_data )
	--LOG('LOAD: Sim side of load running with options: '..repr(query_data)) -- warning large repr logging can crash FA entirely
	LOG('LOAD: Sim side of load running.')
	local save_data = table.deserialise(query_data.Data)
	if save_data.entity_data then
		-- Now we need to replace all entities (except environment) with saved ones
		ClearAllEntities()
		safecall( 'Restoring entities '..repr(save_data.entity_data), 
			CreateEntities, save_data.entity_data )
		-- Reassign unit states and orders
		for old_id, unit_and_data in units do
			safecall( 'Restoring state for unit '..old_id..' data: '..repr(unit_and_data),
				RestoreUnitState, old_id, unit_and_data )
		end
		-- Reattach transported units
		for transport_id, units in transports do
			safecall( 'Attaching units to transport '..transport_id, 
				AttachTransportedUnits, transport_id, units )
		end
	else
		error('LOAD: No entities in save data')
	end
	-- Restore army data
	if save_data.armies_data then
		for army_name, army_data in save_data.armies_data do
			LOG('LOAD: Restoring army '..army_name)
			safecall( 'Restoring army '..army_name, 
				RestoreArmy, army_name, army_data )
		end
	else
		error('LOAD: No armies in save data')	
	end
	-- give back memory
	ClearGlobals()
	return true
end

-- test load and save
function Test( )
	local function SaveReloadThread()
		-- store the data
		SaveGame( )
		-- wait
		WaitSeconds(20)
		-- restore the save game
		LoadGame( )
	end
	ForkThread(SaveReloadThread)
end

-- Add query listeners. These return values to the User scope when a query is sent.
-- Using sim callbacks currently
--SimQuery.AddQueryListener( 'SaveGame', SaveGame )
--SimQuery.AddQueryListener( 'LoadGame', LoadGame )
